/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-LIB.
 *
 *	PDR-LIB is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-LIB is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-LIB.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.solr;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.net.URI;
import java.net.URLEncoder;
import java.util.Map;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.StatusLine;
import org.apache.http.TruncatedChunkException;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.CloseableHttpClient;

/**
 * Takes parameters from incoming HTTP request and builds
 * and URL for a Solr query via HTTP.
 * @author Fabian Körner
 */
public class SolrHttpQuery extends AbstractSolrCommand<String>
{
	private static final String PATH_SELECT = "/select";

	private Map<String, String> _parameters;
	private String _uriPath;

	/**
	 * Forwards the provided query with its parameters to Solr's HTTP interface and returns
	 * the response.
	 * @param parameters Solr parameters included in query.
	 * @throws SolrCommandException Could not successfully create Solr command.
	 */
	public SolrHttpQuery(final Map<String, String> parameters) throws SolrCommandException
	{
		_parameters = parameters;
		_uriPath = SolrConfiguration.getInstance().getBaseUri().toString() + PATH_SELECT;
	}

	/**
	 * @see org.bbaw.pdr.solr.AbstractSolrCommand#execute()
	 * @return Query response.
	 * @throws SolrCommandException Could not successfully execute Solr command.
	 */
	public final String execute() throws SolrCommandException
	{
		try
		{
			URI requestURI = this.buildURIWithParameters();

			CloseableHttpClient client = super.getHttpClientBuilder().build();
			HttpGet get = new HttpGet(requestURI);
			HttpResponse response = client.execute(get);

			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				String result = this.getContent(response);
				client.close();
				return result;
			}
			else
			{
				client.close();
				throw new SolrCommandException("error while querying solr via HTTP; statusCode=" + statusCode);
			}
		}
		catch (Exception cause)
		{
			throw new SolrCommandException(cause);
		}
	}

	private URI buildURIWithParameters() throws SolrCommandException
	{
		int count = 0;
		StringBuilder sb = new StringBuilder(_uriPath);
		for (String key : _parameters.keySet())
		{
			sb.append((count == 0) ? '?' : '&');
			sb.append(key);
			sb.append('=');
			try
			{
				sb.append(URLEncoder.encode(_parameters.get(key), "UTF-8"));
			}
			catch (Exception cause)
			{
				throw new SolrCommandException("could not build URI with parameters", cause);
			}
			count++;
		}
		return URI.create(sb.toString());
	}

	private String getContent(final HttpResponse response) throws SolrCommandException
	{
		String result = null;
		try
		{
			HttpEntity entity = response.getEntity();
			BufferedReader br = new BufferedReader(new InputStreamReader(entity.getContent()));

			StringBuilder sb = new StringBuilder();
			char[] cBuffer = new char[2048];
			int read = 0;
			while ((read = br.read(cBuffer)) != -1)
			{
				sb.append(new String(cBuffer, 0, read));
			}

			result = sb.toString().trim();
		}
		catch (TruncatedChunkException ignored)
		{
			//TODO that's scary .. there must be another way to handle truncated responses, or is it a server problem?
			result = new SolrHttpQuery(_parameters).execute();
		}
		catch (Exception cause)
		{
			throw new SolrCommandException("could not get content for solr http query", cause);
		}
		return result;
	}
}
