/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-LIB.
 *
 *	PDR-LIB is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-LIB is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-LIB.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.utils.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import org.bbaw.pdr.padre.request.PadreID;
import org.bbaw.pdr.utils.Identifier;
import org.bbaw.pdr.utils.PDRType;
import org.bbaw.pdr.utils.ValidationException;
import org.junit.Test;

/**
 * @author Fabian Körner
 */
public class TestIdentifier
{

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testIdentifierFromString() throws Exception
	{
		new Identifier("pdrAo.001.001.000000001");
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testIdentifierFromComponents() throws Exception
	{
		new Identifier(PDRType.USER, 2, 123, 37);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = ValidationException.class)
	public final void testInvalidIdentifierFromStringWithInvalidType() throws Exception
	{
		try
		{
			new Identifier("pdrLo.001.001.000000001");
		}
		catch (Exception cause)
		{
			assertFalse("exeception message must contain pdrLo.001.001.000000001", -1 == cause.getMessage().indexOf("pdrLo.001.001.000000001"));
			throw cause;
		}
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = ValidationException.class)
	public final void testInvalidIdentifierFromStringWithInvalidInstance() throws Exception
	{
		try
		{
			new Identifier("pdrUo.2.123.000000037");
		}
		catch (Exception cause)
		{
			assertFalse("exeception message must contain pdrUo.2.123.000000037", -1 == cause.getMessage().indexOf("pdrUo.2.123.000000037"));
			throw cause;
		}
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = ValidationException.class)
	public final void testInvalidIdentifierFromStringWithInvalidProject() throws Exception
	{
		try
		{
			new Identifier("pdrUo.002.1234.000000037");
		}
		catch (Exception cause)
		{
			assertFalse("exeception message must contain pdrUo.002.1234.000000037", -1 == cause.getMessage().indexOf("pdrUo.002.1234.000000037"));
			throw cause;
		}
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = ValidationException.class)
	public final void testInvalidIdentifierFromStringWithInvalidObject() throws Exception
	{
		try
		{
			new Identifier("pdrUo.002.008.0000000f7");
		}
		catch (Exception cause)
		{
			assertFalse("exeception message must contain pdrUo.002.008.0000000f7", -1 == cause.getMessage().indexOf("pdrUo.002.008.0000000f7"));
			throw cause;
		}
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = ValidationException.class)
	public final void testInvalidIdentifierFromComponentsWithInvalidInstance() throws Exception
	{
		try
		{
			new Identifier(PDRType.USER, 0, 123, 37);
		}
		catch (Exception cause)
		{
			assertFalse("exeception message must contain pdrUo.000.123.000000037", -1 == cause.getMessage().indexOf("pdrUo.000.123.000000037"));
			throw cause;
		}
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = ValidationException.class)
	public final void testInvalidIdentifierFromComponentsWithInvalidProject() throws Exception
	{
		try
		{
			new Identifier("pdrUo.002.1234.000000037");
		}
		catch (Exception cause)
		{
			assertFalse("exeception message must contain pdrUo.002.1234.000000037", -1 == cause.getMessage().indexOf("pdrUo.002.1234.000000037"));
			throw cause;
		}
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = ValidationException.class)
	public final void testInvalidIdentifierFromComponentsWithInvalidObject() throws Exception
	{
		try
		{
			new Identifier("pdrUo.002.134.200000000");
		}
		catch (Exception cause)
		{
			assertFalse("exeception message must contain pdrUo.002.134.200000000", -1 == cause.getMessage().indexOf("pdrUo.002.134.200000000"));
			throw cause;
		}
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = ValidationException.class)
	public final void testInvalidIdentifierFromStringWithZero() throws Exception
	{
		new Identifier("pdrAo.000.001.000000001");
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = ValidationException.class)
	public final void testInvalidIdentifierFromComponentsWithZero() throws Exception
	{
		new Identifier(PDRType.USER, 2, 0, 0);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = ValidationException.class)
	public final void testInvalidTemporaryIdentifierWithObjectZero() throws Exception
	{
		new Identifier(PDRType.ASPECT, 2, 3, 100000000);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testTypeFromString() throws Exception
	{
		for (PDRType t : PDRType.values())
		{
			assertEquals(t, Identifier.typeFromString(t.toString()));
		}
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testEqualsFalseProject() throws Exception
	{
		Identifier id0 = new Identifier("pdrAo.100.111.110100110");
		Identifier id1 = new Identifier(PDRType.ASPECT, 100, 123, 110100110);
		assertFalse(id0.equals(id1));
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testGetComponents() throws Exception
	{
		Identifier id = new Identifier("pdrRo.123.456.012345678");
		assertEquals(PDRType.REFERENCE, id.getType());
		assertEquals(123, id.getInstanceNumber());
		assertEquals(456, id.getProjectNumber());
		assertEquals(12345678, id.getObjectNumber());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testCreateReturnsInstance() throws Exception
	{
		Identifier id = Identifier.create("pdrUo.004.999.000004444");
		assertFalse(null == id);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testCreateReturnsNull() throws Exception
	{
		Identifier id = Identifier.create("pdrUo.4.999.000004444");
		assertTrue(null == id);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testToPadreID() throws Exception
	{
		Identifier id = Identifier.create("pdrPo.015.798.000004444");
		PadreID result = id.toPadreID();
		assertEquals("pdrPo", result.getPartType());
		assertEquals("00f", result.getPartInstance());
		assertEquals("31e", result.getPartStore());
		assertEquals("00000115c", result.getPartObject());

		assertEquals(15, result.getValueInstance());
		assertEquals(798, result.getValueStore());
		assertEquals(4444, result.getValueObject());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testIntToHexString0() throws Exception
	{
		String result = Identifier.intToHexString(0, 16);
		assertEquals("0000000000000000", result);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testIntToHexString10() throws Exception
	{
		String result = Identifier.intToHexString(10, 5);
		assertEquals("0000a", result);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testIntToHexString255() throws Exception
	{
		String result = Identifier.intToHexString(255, 2);
		assertEquals("ff", result);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testIntToHexString435786() throws Exception
	{
		String result = Identifier.intToHexString(435786, 10);
		assertEquals("000006a64a", result);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testIntToHexStringLength1() throws Exception
	{
		String result = Identifier.intToHexString(10, 1);
		assertEquals("a", result);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = IllegalArgumentException.class)
	public final void testIntToHexStringResultTooLong() throws Exception
	{
		Identifier.intToHexString(435786, 3);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = IllegalArgumentException.class)
	public final void testIntToHexStringLength0() throws Exception
	{
		String result = Identifier.intToHexString(123, 0);
		assertEquals("000006a64a", result);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = IllegalArgumentException.class)
	public final void testIntToHexStringLengthNegative() throws Exception
	{
		String result = Identifier.intToHexString(123, -4);
		assertEquals("000006a64a", result);
	}
}
