/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-LIB.
 *
 *	PDR-LIB is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-LIB is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-LIB.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.utils.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.solr.client.solrj.SolrServer;
import org.apache.solr.client.solrj.impl.HttpSolrServer;
import org.apache.solr.client.solrj.response.SolrPingResponse;
import org.apache.solr.common.SolrInputDocument;
import org.bbaw.pdr.padre.PadreConfiguration;
import org.bbaw.pdr.padre.client.PadreClient;
import org.bbaw.pdr.solr.SolrConfiguration;
import org.bbaw.pdr.utils.IDRange;
import org.bbaw.pdr.utils.Identifier;
import org.bbaw.pdr.utils.PDRObject;
import org.bbaw.pdr.utils.PDRType;
import org.bbaw.pdr.utils.Repository;
import org.bbaw.pdr.utils.RepositoryException;
import org.bbaw.pdr.utils.ValidationException;
import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

/**
* @author Fabian Körner
*/
public class TestRepository
{

	private static final TestConfiguration CONFIGURATION = new TestConfigurationFromFile();

	private static SolrServer _solrServer;
	private static PadreClient _padreClient;

	/**
	 * @throws Exception Exception.
	 */
	@BeforeClass
	public static void setUp() throws Exception
	{
		SolrConfiguration.getInstance().setBaseURI(CONFIGURATION.getSolrUri());
		_solrServer = SolrConfiguration.getInstance().getSolrServer();

		PadreConfiguration.getInstance().setCredentials(CONFIGURATION.getPadreUri(), CONFIGURATION.getPadreUsername(), CONFIGURATION.getPadrePassword());
		_padreClient = PadreConfiguration.getInstance().getPadreClient();

		Collection<SolrInputDocument> inputDocuments = new Vector<SolrInputDocument>();
		inputDocuments.add(TestRepository.createSimpleAspect());
		inputDocuments.add(TestRepository.createPersonWithPND());
		inputDocuments.add(TestRepository.createPersonWithTwoExternalIDs());
		inputDocuments.add(TestRepository.createSimpleUser());
		_solrServer.add(inputDocuments);
		_solrServer.commit();

		List<String> objects = new Vector<String>();
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.790.100000001"));
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.790.100000002"));
		Repository.ingestObjects(1, 790, objects);

		objects.clear();
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.791.100000001"));
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.791.100000002"));
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.791.100000003"));
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.791.100000004"));
		Repository.ingestObjects(1, 791, objects);

		objects.clear();
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.793.100000001"));
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.793.100000002"));
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.793.100000003"));
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.793.100000004"));
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.793.100000005"));
		Repository.ingestObjects(1, 793, objects);
	}

	/**
	 * @throws Exception Exception.
	 */
	@AfterClass
	public static final void tearDown() throws Exception
	{
		_solrServer.deleteByQuery("id.project:789");
		_solrServer.deleteByQuery("id.project:790");
		_solrServer.deleteByQuery("id.project:791");
		_solrServer.deleteByQuery("id.project:792");
		_solrServer.deleteByQuery("id.project:793");
		_solrServer.commit();

		((HttpSolrServer) _solrServer).shutdown();

		/* 315 doesn't exist */
		_padreClient.deleteStore("316");
		_padreClient.deleteStore("317");
		_padreClient.deleteStore("318");
		_padreClient.deleteStore("319");
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testPing() throws Exception
	{
		SolrPingResponse response = _solrServer.ping();
		assertEquals(0, response.getStatus());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testGetUserID() throws Exception
	{
		Identifier id = Repository.getUserID("tester", 789);
		String expected = "pdrUo.001.789.000000023";
		String actual = id.toString();
		assertEquals(expected, actual);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testGetProjectSizes() throws Exception
	{
		Map<Integer, Integer> result = Repository.getProjectSizes();
		assertNotNull("at least an empty map should be returned", result);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test(expected = RepositoryException.class)
	public final void testGetUserIDWithUnknownUsername() throws Exception
	{
		Repository.getUserID("unknown", 789);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test(expected = RepositoryException.class)
	public final void testGetUserIDWithWrongProject() throws Exception
	{
		Repository.getUserID("tester", 987);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testDeleteIDRange() throws Exception
	{
		String expected = "[IDRange[pdrPo,1,793,1-5]]";
		String actual = Repository.getOccupiedObjectIDRanges(PDRType.PERSON, 1, 793, 1, 5).toString();
		assertEquals("precondition failed", expected, actual);

		Repository.deleteIDRange(PDRType.PERSON, 1, 793, 1, 3);

		expected = "[IDRange[pdrPo,1,793,4-5]]";
		actual = Repository.getOccupiedObjectIDRanges(PDRType.PERSON, 1, 793, 1, 5).toString();
		assertEquals("one range with persons 4 and 5 should be returned", expected, actual);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testIngestObjectsInExistingProject() throws Exception
	{
		List<IDRange> rangesBefore = Repository.getOccupiedObjectIDRanges(PDRType.PERSON, 1, 790, 1, 9999);
		assertTrue(1 == rangesBefore.size());
		assertTrue(2 == rangesBefore.get(0).count());

		List<String> objects = new Vector<String>();
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.790.100000001"));
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.790.100000002"));
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.790.100000004"));
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.790.100000003"));
		Map<Identifier, Identifier> result = Repository.ingestObjects(1, 790, objects);

		assertTrue(result.containsValue(new Identifier("pdrPo.001.790.000000003")));
		assertTrue(result.containsValue(new Identifier("pdrPo.001.790.000000005")));
		assertTrue(result.containsValue(new Identifier("pdrPo.001.790.000000004")));
		assertTrue(result.containsValue(new Identifier("pdrPo.001.790.000000006")));

		List<IDRange> rangesAfter = Repository.getOccupiedObjectIDRanges(PDRType.PERSON, 1, 790, 1, 9999);
		assertEquals(1, rangesAfter.size());
		assertEquals(6, rangesAfter.get(0).count());
		assertEquals(1, rangesAfter.get(0).getLowerBound());
		assertEquals(6, rangesAfter.get(0).getUpperBound());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testIngestObjectsInNewProject() throws Exception
	{
		List<String> objects = new Vector<String>();
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.792.100000001"));
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.792.100000002"));
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.792.100000004"));
		objects.add(TestRepository.createNewPersonObject("pdrPo.001.792.100000003"));
		Map<Identifier, Identifier> result = Repository.ingestObjects(1, 792, objects);

		assertTrue(result.containsValue(new Identifier("pdrPo.001.792.000000001")));
		assertTrue(result.containsValue(new Identifier("pdrPo.001.792.000000002")));
		assertTrue(result.containsValue(new Identifier("pdrPo.001.792.000000003")));
		assertTrue(result.containsValue(new Identifier("pdrPo.001.792.000000004")));

		List<IDRange> ranges = Repository.getOccupiedObjectIDRanges(PDRType.PERSON, 1, 792, 1, 9999);
		assertEquals(1, ranges.size());
		assertEquals(4, ranges.get(0).count());
		assertEquals(1, ranges.get(0).getLowerBound());
		assertEquals(4, ranges.get(0).getUpperBound());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testModifyObjects() throws Exception
	{
		Thread.sleep(1000L);
		String timestamp = Repository.getTime();
		Thread.sleep(1000L);

		List<String> objects = new Vector<String>();
		objects.add(TestRepository.createModifiedPersonObject("pdrPo.001.791.000000003"));
		objects.add(TestRepository.createModifiedPersonObject("pdrPo.001.791.000000004"));

		Repository.modifyObjects(1, 791, objects, false);

		List<String> result = Repository.getModifiedObjects(1, 791, timestamp);
		assertEquals(2, result.size());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = ValidationException.class)
	public final void testModifyInvalidObjects() throws Exception
	{
		Thread.sleep(1000L);
		String timestamp = Repository.getTime();
		Thread.sleep(1000L);

		List<String> objects = new Vector<String>();
		objects.add(TestRepository.createModifiedPersonObject("pdrPo.001.791.000000003").replaceAll("ref=", "bla="));
		objects.add(TestRepository.createModifiedPersonObject("pdrPo.001.791.000000004").replaceAll("ref=", "bla="));

		Repository.modifyObjects(1, 791, objects, false);

		List<String> result = Repository.getModifiedObjects(1, 791, timestamp);
		assertEquals(0, result.size());
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = RepositoryException.class)
	public final void testModifyObjectsInNonExistentProject() throws Exception
	{

		List<String> objects = new Vector<String>();
		objects.add(TestRepository.createModifiedPersonObject("pdrPo.001.793.000000503"));
		objects.add(TestRepository.createModifiedPersonObject("pdrPo.001.793.000000504"));

		Repository.modifyObjects(1, 793, objects, false);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = RepositoryException.class)
	public final void testModifyObjectsInNonExistentProjectWithForce() throws Exception
	{

		List<String> objects = new Vector<String>();
		objects.add(TestRepository.createModifiedPersonObject("pdrPo.001.793.000000503"));
		objects.add(TestRepository.createModifiedPersonObject("pdrPo.001.793.000000504"));

		Repository.modifyObjects(1, 777, objects, true);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testGetObjects() throws Exception
	{
		List<String> results = Repository.getObjects(PDRType.PERSON, 1, 790, 1, 2);
		assertEquals(2, results.size());
		Pattern pattern = Pattern.compile("^<person\\s.+</person>$");
		Matcher matcher;
		for (String result : results)
		{
			matcher = pattern.matcher(result);
			assertTrue(matcher.find());
		}
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test
	public final void testCheckZeroRemainingTemporaryIDs() throws Exception
	{

		String tempReferenceID = "<aspect xmlns=\"http://pdr.bbaw.de/namespaces/aodl/\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocat"
				+ "ion=\"http://pdr.bbaw.de/namespaces/aodl/ http://pdr.bbaw.de/schema/aodl.xsd\" id=\"pdrAo.001.512.000000001\"><record><revision ref=\"0\" ti"
				+ "mestamp=\"2014-02-12T09:59:50\" authority=\"pdrUo.001.512.000000001\"/></record><timeDim><timeStm type=\"undefined\"/></timeDim><spatialDim>"
				+ "<spatialStm type=\"undefined\"/></spatialDim><relationDim><relationStm subject=\"pdrAo.001.512.000000001\"><relation object=\"pdrPo.001.512."
				+ "000000001\" provider=\"PDR\">aspect_of</relation></relationStm></relationDim><semanticDim><semanticStm provider=\"PDR_2.0\">NormName_DE</sem"
				+ "anticStm></semanticDim><notification><persName type=\"forename\" ana=\"pdrPo.001.512.000000001\">Maxi</persName> <persName type=\"surname\" "
				+ "ana=\"pdrPo.001.512.000000001\">Manna</persName></notification><validation><validationStm authority=\"pdrUo.001.512.000000001\"><reference q"
				+ "uality=\"certain\">pdrRo.001.002.000000001</reference><interpretation/></validationStm></validation></aspect>";

		List<PDRObject> objects = new Vector<PDRObject>();
		objects.add(new PDRObject(tempReferenceID));

		Repository.checkRemainingTemporaryIDs(objects);
	}

	/**
	 * @throws Exception Exception.
	 */
	@Test (expected = RepositoryException.class)
	public final void testCheckOneRemainingTemporaryIDs() throws Exception
	{

		String tempReferenceID = "<aspect xmlns=\"http://pdr.bbaw.de/namespaces/aodl/\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:schemaLocat"
				+ "ion=\"http://pdr.bbaw.de/namespaces/aodl/ http://pdr.bbaw.de/schema/aodl.xsd\" id=\"pdrAo.001.512.000000001\"><record><revision ref=\"0\" ti"
				+ "mestamp=\"2014-02-12T09:59:50\" authority=\"pdrUo.001.512.000000001\"/></record><timeDim><timeStm type=\"undefined\"/></timeDim><spatialDim>"
				+ "<spatialStm type=\"undefined\"/></spatialDim><relationDim><relationStm subject=\"pdrAo.001.512.000000001\"><relation object=\"pdrPo.001.512."
				+ "000000001\" provider=\"PDR\">aspect_of</relation></relationStm></relationDim><semanticDim><semanticStm provider=\"PDR_2.0\">NormName_DE</sem"
				+ "anticStm></semanticDim><notification><persName type=\"forename\" ana=\"pdrPo.001.512.000000001\">Maxi</persName> <persName type=\"surname\" "
				+ "ana=\"pdrPo.001.512.000000001\">Manna</persName></notification><validation><validationStm authority=\"pdrUo.001.512.000000001\"><reference q"
				+ "uality=\"certain\">pdrRo.001.002.100000001</reference><interpretation/></validationStm></validation></aspect>";

		List<PDRObject> objects = new Vector<PDRObject>();
		objects.add(new PDRObject(tempReferenceID));

		Repository.checkRemainingTemporaryIDs(objects);
	}

	private static SolrInputDocument createSimpleAspect()
	{
		SolrInputDocument result = new SolrInputDocument();
		result.addField("id", "pdrAo.001.789.000002979");
		result.addField("id.type", "pdrAo");
		result.addField("id.instance", 1);
		result.addField("id.project", 789);
		result.addField("id.object", 2979);
		result.addField("revision.timestamp", "2012-01-01T12:00:00Z");
		result.addField("notification", "narf");
		return result;
	}

	private static SolrInputDocument createPersonWithPND()
	{
		SolrInputDocument result = new SolrInputDocument();
		result.addField("id", "pdrPo.001.789.000002972");
		result.addField("id.type", "pdrPo");
		result.addField("id.instance", 1);
		result.addField("id.project", 789);
		result.addField("id.object", 2972);
		result.addField("revision.timestamp", "2012-01-01T13:00:00Z");
		result.addField("person.identifier", "123456789X");
		result.addField("person.identifier.provider", "DNB");
		result.addField("person.cidentifier", "123456789X|DNB");
		return result;
	}

	private static SolrInputDocument createPersonWithTwoExternalIDs()
	{
		SolrInputDocument result = new SolrInputDocument();
		result.addField("id", "pdrPo.001.789.000002973");
		result.addField("id.type", "pdrPo");
		result.addField("id.instance", 1);
		result.addField("id.project", 789);
		result.addField("id.object", 2973);
		result.addField("revision.timestamp", "2012-01-01T14:00:00Z");
		result.addField("person.identifier", "123456789X");
		result.addField("person.identifier.provider", "DNB");
		result.addField("person.cidentifier", "123456789X|DNB");
		result.addField("person.identifier", "100340");
		result.addField("person.identifier.provider", "nobody");
		result.addField("person.cidentifier", "100340|nobody");
		return result;
	}

	private static SolrInputDocument createSimpleUser()
	{
		SolrInputDocument result = new SolrInputDocument();
		result.addField("id", "pdrUo.001.789.000000023");
		result.addField("id.type", "pdrUo");
		result.addField("id.instance", 1);
		result.addField("id.project", 789);
		result.addField("id.object", 23);
		result.addField("user.username", "tester");
		return result;
	}

	private static String createNewPersonObject(final String id)
	{
		return new String("<person xmlns=\"http://pdr.bbaw.de/namespaces/podl/\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\""
				+ " xsi:schemaLocation=\"http://pdr.bbaw.de/namespaces/podl/ http://pdr.bbaw.de/schema/podl.xsd\" id=\"" + id + "\">"
				+ "<record><revision ref=\"0\" timestamp=\"2012-11-30T04:48:08\" authority=\"pdrUo.002.762.000000002\"/></record></person>");
	}

	private static String createModifiedPersonObject(final String id)
	{
		return new String("<person xmlns=\"http://pdr.bbaw.de/namespaces/podl/\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\""
				+ " xsi:schemaLocation=\"http://pdr.bbaw.de/namespaces/podl/ http://pdr.bbaw.de/schema/podl.xsd\" id=\"" + id + "\">"
				+ "<record><revision ref=\"0\" timestamp=\"2012-11-30T04:48:08\" authority=\"pdrUo.002.762.000000002\"/>"
				+ "<revision ref=\"1\" timestamp=\"2012-11-30T04:50:08\" authority=\"pdrUo.002.762.000000002\"/></record></person>");
	}
}
