/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PaDRe.
 *
 *	PaDRe is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PaDRe is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PaDRe.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.padre.client;

import java.net.URI;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.Properties;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.StatusLine;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.CredentialsProvider;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.protocol.HttpClientContext;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.BasicCredentialsProvider;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.protocol.BasicHttpContext;
import org.apache.http.protocol.HttpContext;
import org.apache.http.util.EntityUtils;
import org.bbaw.pdr.padre.request.PadreID;
import org.bbaw.pdr.padre.request.PadreObject;

/**
 * Single class implementation of a PaDRe client.
 * @author Fabian Körner
 */
public class PadreClient
{

	private static final String PATH_DISCOVER = "/discover";

	private HttpClientBuilder _httpClientBuilder = HttpClientBuilder.create();
	private HttpContext _httpContext;
	private CredentialsProvider _credentialsProvider;

	private URI _padreAPI;
	private Properties _methodPaths = new Properties();

	/**
	 * Constructor with information on how to get connected to PaDRe's web API.
	 * @param padreAPI Base URI of PaDRe's web API.
	 * @param username Username for Authentication.
	 * @param password Password for Authentication.
	 * @throws PadreClientException Could not get connection to PaDRe API.
	 */
	public PadreClient(final URI padreAPI, final String username, final String password) throws PadreClientException
	{
		_padreAPI = padreAPI;

		_httpContext = new BasicHttpContext();

		_credentialsProvider = new BasicCredentialsProvider();
		_credentialsProvider.setCredentials(
				    new AuthScope(padreAPI.getHost(), padreAPI.getPort()),
				    new UsernamePasswordCredentials(username, password));

		_httpContext.setAttribute(HttpClientContext.CREDS_PROVIDER, _credentialsProvider);

		this.discover();
	}

	/**
	 * Create a store in the database back-end.
	 * @param storeID Store ID (hex).
	 * @throws PadreClientException Could not create the store.
	 */
	public final void createStore(final String storeID) throws PadreClientException
	{
		URI requestURI = this.createRequestURI("GET_CREATE_STORE", storeID);

		try
		{
			CloseableHttpClient client = _httpClientBuilder.build();
			HttpGet get = new HttpGet(requestURI);

			HttpResponse response = client.execute(get, _httpContext);
			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				client.close();
			}
			else
			{
				client.close();
				throw new PadreClientException("createStore: failed to create store [" + requestURI.toString() + "]; http status code: " + statusCode);
			}
		}
		catch (Throwable cause)
		{
			throw new PadreClientException("createStore: failed to create store [" + requestURI.toString() + "]", cause);
		}
	}

	/**
	 * Delete a store and its content from database back-end.
	 * @param storeID Store ID (hex).
	 * @throws PadreClientException Could not delete the store.
	 */
	public final void deleteStore(final String storeID) throws PadreClientException
	{
		URI requestURI = this.createRequestURI("DELETE_DELETE_STORE", storeID);

		try
		{
			CloseableHttpClient client = _httpClientBuilder.build();

			HttpDelete delete = new HttpDelete(requestURI);

			HttpResponse response = client.execute(delete, _httpContext);
			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				client.close();
			}
			else
			{
				client.close();
				throw new PadreClientException("deleteStore: failed to delete store [" + requestURI.toString() + "]; http status code: " + statusCode);
			}
		}
		catch (Throwable cause)
		{
			throw new PadreClientException("deleteStore: failed to delete store [" + requestURI.toString() + "]", cause);
		}
	}

	/**
	 * Retrieves all available IDs and corresponding formats for all methods
	 * of PaDRe's API.
	 * @throws PadreClientException Fatal: Could not discover methods of PaDRe's API.
	 */
	private void discover() throws PadreClientException
	{
		URI requestURI = URI.create(_padreAPI.toString() + PATH_DISCOVER);

		try
		{
			CloseableHttpClient client = _httpClientBuilder.build();
			HttpGet get = new HttpGet(requestURI);

			HttpResponse response = client.execute(get, _httpContext);
			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				HttpEntity responseEntity = response.getEntity();
				_methodPaths.load(responseEntity.getContent());
				client.close();
			}
			else
			{
				client.close();
				throw new PadreClientException("discover: failed [" + requestURI.toString() + "]; http status code: " + statusCode);
			}
		}
		catch (Throwable cause)
		{
			throw new PadreClientException("discover: failed [" + requestURI.toString() + "]", cause);
		}
	}

	/**
	 * Returns all method IDs and their formats as <code>java.util.Properties</code>.
	 * @return Map from method IDs to formats.
	 */
	public final Properties getMethodPaths()
	{
		return _methodPaths;
	}

	/**
	 * Try to do an optimisation of the provided store in the database back-end.
	 * @param storeID Store ID (hex).
	 * @throws PadreClientException Could not optimise the store.
	 */
	public final void optimizeStore(final String storeID) throws PadreClientException
	{
		URI requestURI = this.createRequestURI("GET_OPTIMIZE_STORE", storeID);

		try
		{
			CloseableHttpClient client = _httpClientBuilder.build();
			HttpGet get = new HttpGet(requestURI);

			HttpResponse response = client.execute(get, _httpContext);
			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				client.close();
			}
			else
			{
				client.close();
				throw new PadreClientException("optimizeStore: failed [" + requestURI.toString() + "]; http status code: " + statusCode);
			}
		}
		catch (Throwable cause)
		{
			throw new PadreClientException("optimizeStore: failed [" + requestURI.toString() + "]" + storeID, cause);
		}
	}

	/**
	 * Checks the availability of PaDRe by trying to retrieve an empty response.
	 * @return PaDRe's API is accessable.
	 * @throws PadreClientException Ping not successful.
	 */
	public final boolean ping() throws PadreClientException
	{
		URI requestURI = this.createRequestURI("GET_PING");
		boolean result = true;

		try
		{
			CloseableHttpClient client = _httpClientBuilder.build();
			HttpGet get = new HttpGet(requestURI);

			HttpResponse response = client.execute(get, _httpContext);
			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				HttpEntity responseEntity = response.getEntity();
				String content  = EntityUtils.toString(responseEntity);
				result = 0 == content.length();
				client.close();
			}
			else
			{
				client.close();
				throw new PadreClientException("ping: failed [" + requestURI.toString() + "]; http status code: " + statusCode);
			}
		}
		catch (Throwable cause)
		{
			throw new PadreClientException("ping: failed [" + requestURI.toString() + "]", cause);
		}
		return result;
	}

	/**
	 * Delete the content of the provided store.
	 * @param storeID Store ID (hex).
	 * @throws PadreClientException Could nto reset store.
	 */
	public final void resetStore(final String storeID) throws PadreClientException
	{
		URI requestURI = this.createRequestURI("GET_RESET_STORE", storeID);

		try
		{
			CloseableHttpClient client = _httpClientBuilder.build();
			HttpGet get = new HttpGet(requestURI);

			HttpResponse response = client.execute(get, _httpContext);
			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				client.close();
			}
			else
			{
				client.close();
				throw new PadreClientException("resetStore: failed [" + requestURI.toString() + "]; http status code: " + statusCode);
			}
		}
		catch (Throwable cause)
		{
			throw new PadreClientException("resetStore: failed [" + requestURI.toString() + "]", cause);
		}
	}

	/**
	 * Returns the current server time of PaDRe's host.
	 * @return Server time.
	 * @throws PadreClientException Could not get server time.
	 */
	public final String time() throws PadreClientException
	{
		URI requestURI = this.createRequestURI("GET_TIME");
		String result;

		try
		{
			CloseableHttpClient client = _httpClientBuilder.build();
			HttpGet get = new HttpGet(requestURI);

			HttpResponse response = client.execute(get, _httpContext);
			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				HttpEntity responseEntity = response.getEntity();
				result = EntityUtils.toString(responseEntity);
				client.close();
			}
			else
			{
				client.close();
				throw new PadreClientException("time: failed [" + requestURI.toString() + "]; http status code: " + statusCode);
			}
		}
		catch (Throwable cause)
		{
			throw new PadreClientException("time: failed [" + requestURI.toString() + "]", cause);
		}
		return result;
	}

	/**
	 * Writes provided PaDRe objects to the database back-end.
	 * @param padreObjects PaDRe objects.
	 * @throws PadreClientException Could not write to database back-end.
	 */
	public final void write(final List<PadreObject> padreObjects) throws PadreClientException
	{
		URI requestURI = this.createRequestURI("PUT_WRITE");

		try
		{
			CloseableHttpClient client = _httpClientBuilder.build();
			HttpPut put = new HttpPut(requestURI);

			String content = padreObjects.toString();
			put.setEntity(new StringEntity(content, "UTF-8"));

			HttpResponse response = client.execute(put, _httpContext);
			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				client.close();
			}
			else
			{
				client.close();
				throw new PadreClientException("write: failed to write objects [" + requestURI.toString() + "]; http status code: " + statusCode);
			}
		}
		catch (Throwable cause)
		{
			throw new PadreClientException("write: failed to write objects [" + requestURI.toString() + "]", cause);
		}
	}

	/**
	 * Reads PaDRe objects from database back-end.
	 * @param padreIDs List of PaDRe IDs.
	 * @return PaDRe objects.
	 * @throws PadreClientException Could not read objects from database back-end.
	 */
	public final List<PadreObject> read(final List<PadreID> padreIDs) throws PadreClientException
	{
		URI requestURI = this.createRequestURI("POST_READ");
		final List<PadreObject> result = new Vector<PadreObject>();

		try
		{
			CloseableHttpClient client = _httpClientBuilder.build();
			HttpPost post = new HttpPost(requestURI);

			String content = padreIDs.toString();
			post.setEntity(new StringEntity(content, "UTF-8"));

			HttpResponse response = client.execute(post, _httpContext);
			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				String responseContent = EntityUtils.toString(response.getEntity());
				Pattern pattern = PadreObject.PATTERN_COMBINED;
				Matcher matcher = pattern.matcher(responseContent);
				String combined;
				while (matcher.find())
				{
					combined = matcher.group();
					result.add(new PadreObject(combined));
				}
				client.close();
			}
			else
			{
				client.close();
				throw new PadreClientException("read: failed to read objects [" + requestURI.toString() + "]; http status code: " + statusCode);
			}
		}
		catch (Throwable cause)
		{
			throw new PadreClientException("read: failed to read objects [" + requestURI.toString() + "]", cause);
		}
		return result;
	}

	/**
	 * Deletes PaDRe objects from database back-end according to the provided list.
	 * @param padreIDs IDs of objects which will be deleted.
	 * @throws PadreClientException Could not delete objects.
	 */
	public final void delete(final List<PadreID> padreIDs) throws PadreClientException
	{
		URI requestURI = this.createRequestURI("POST_DELETE");

		try
		{
			CloseableHttpClient client = _httpClientBuilder.build();
			HttpPost post = new HttpPost(requestURI);

			String content = padreIDs.toString();
			post.setEntity(new StringEntity(content, "UTF-8"));

			HttpResponse response = client.execute(post, _httpContext);
			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				client.close();
			}
			else
			{
				client.close();
				throw new PadreClientException("delete: failed to delete objects [" + requestURI.toString() + "]; http status code: " + statusCode);
			}
		}
		catch (Throwable cause)
		{
			throw new PadreClientException("delete: failed to delete objects [" + requestURI.toString() + "]", cause);
		}
	}

	/**
	 * Returns a list of available stores.
	 * @return List of stores.
	 * @throws PadreClientException Could not get list of stores.
	 */
	public final List<String> stores() throws PadreClientException
	{
		URI requestURI = this.createRequestURI("GET_STORES");
		final List<String> result = new Vector<String>();

		try
		{
			CloseableHttpClient client = _httpClientBuilder.build();
			HttpGet get = new HttpGet(requestURI);

			HttpResponse response = client.execute(get, _httpContext);
			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				HttpEntity responseEntity = response.getEntity();
				String content = EntityUtils.toString(responseEntity);
				Pattern pattern = Pattern.compile("[0-9a-fA-F]{3}");
				Matcher matcher = pattern.matcher(content);
				while (matcher.find())
				{
					result.add(matcher.group());
				}
				client.close();
			}
			else
			{
				client.close();
				throw new PadreClientException("stores: failed to get store IDs [" + requestURI.toString() + "]; http status code: " + statusCode);
			}
		}
		catch (Throwable cause)
		{
			throw new PadreClientException("stores: failed to get store IDs [" + requestURI.toString() + "]", cause);
		}
		return result;
	}

	/**
	 * Checks if an object with the provided ID exists in the database back-end.
	 * @param id ID of PaDRe object.
	 * @return PaDRe object exists.
	 * @throws PadreClientException Could not check if object exists.
	 */
	public final boolean contains(final PadreID id) throws PadreClientException
	{
		URI requestURI = this.createRequestURI("GET_CONTAINS", id.toString());
		boolean result = false;

		try
		{
			CloseableHttpClient client = _httpClientBuilder.build();
			HttpGet get = new HttpGet(requestURI);

			HttpResponse response = client.execute(get, _httpContext);
			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				HttpEntity responseEntity = response.getEntity();
				String responseContent = EntityUtils.toString(responseEntity);
				result = Boolean.parseBoolean(responseContent);
				client.close();
			}
			else
			{
				client.close();
				throw new PadreClientException("contains: failed to check [" + requestURI.toString() + "]; http status code: " + statusCode);
			}
		}
		catch (Throwable cause)
		{
			throw new PadreClientException("contains: failed to check [" + requestURI.toString() + "]", cause);
		}
		return result;
	}

	/**
	 * Returns the ID of the PaDRe instance.
	 * @return Instance ID.
	 * @throws PadreClientException Could not get the instance ID.
	 */
	public final String instance() throws PadreClientException
	{
		URI requestURI = this.createRequestURI("GET_INSTANCE");
		String result = null;

		try
		{
			CloseableHttpClient client = _httpClientBuilder.build();
			HttpGet get = new HttpGet(requestURI);

			HttpResponse response = client.execute(get, _httpContext);
			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				HttpEntity responseEntity = response.getEntity();
				result = EntityUtils.toString(responseEntity);
				result = result.trim();
				client.close();
			}
			else
			{
				client.close();
				throw new PadreClientException("instance: failed to get ID [" + requestURI.toString() + "]; http status code: " + statusCode);
			}
		}
		catch (Throwable cause)
		{
			throw new PadreClientException("instance: failed to get ID [" + requestURI.toString() + "]", cause);
		}
		return result;
	}

	/**
	 * Returns all available ID ranges of objects of provided type in the provided store.
	 * @param storeID Store ID (hex).
	 * @param type Object type.
	 * @return List of ID ranges.
	 * @throws PadreClientException Could not get ID ranges.
	 */
	public final List<PadreID> ranges(final String storeID, final String type) throws PadreClientException
	{
		URI requestURI = this.createRequestURI("GET_RANGES", storeID, type);
		List<PadreID> result;

		try
		{
			CloseableHttpClient client = _httpClientBuilder.build();
			HttpGet get = new HttpGet(requestURI);

			HttpResponse response = client.execute(get, _httpContext);
			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				result = PadreClient.getPadreIDsFromHttpResponse(response);
				client.close();
			}
			else
			{
				client.close();
				throw new PadreClientException("ranges: failed to get ID ranges [" + requestURI.toString() + "]; http status code: " + statusCode);
			}
			return result;
		}
		catch (Throwable cause)
		{
			throw new PadreClientException("ranges: failed to get ID ranges [" + requestURI.toString() + "]", cause);
		}
	}

	/**
	 * Returns IDs of objects in the given PaDRe store which were modified since the provided timestamp.
	 * @param storeID Store ID (hex).
	 * @param type Object type.
	 * @param since Timestamp.
	 * @return List of IDs of modified objects.
	 * @throws PadreClientException Could not get IDs of modified objects.
	 */
	public final List<PadreID> modified(final String storeID, final String type, final Date since) throws PadreClientException
	{
		URI requestURI = this.createRequestURI("POST_MODIFIED", storeID, type);
		List<PadreID> result;

		try
		{
			CloseableHttpClient client = _httpClientBuilder.build();
			HttpPost post = new HttpPost(requestURI);

			SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
			String content = sdf.format(since);
			post.setEntity(new StringEntity(content, "UTF-8"));

			HttpResponse response = client.execute(post, _httpContext);
			StatusLine statusLine = response.getStatusLine();
			int statusCode = statusLine.getStatusCode();
			if (HttpStatus.SC_OK == statusCode)
			{
				result = PadreClient.getPadreIDsFromHttpResponse(response);

				client.close();
			}
			else
			{
				client.close();
				throw new PadreClientException("modified: failed to get ID ranges [" + requestURI.toString() + "]; http status code: " + statusCode);
			}
			return result;
		}
		catch (Throwable cause)
		{
			throw new PadreClientException("modified: failed to get ID ranges [" + requestURI.toString() + "]", cause);
		}
	}

	private static List<PadreID> getPadreIDsFromHttpResponse(final HttpResponse response) throws Exception
	{
		final List<PadreID> result = new Vector<PadreID>();

		HttpEntity responseEntity = response.getEntity();
		String responseContent = EntityUtils.toString(responseEntity);

		String[] parts = responseContent.split(",");
		for (String part : parts)
		{
			part = part.replaceAll("(\\[|\\])", "").trim();
			if (!part.isEmpty())
			{
				result.add(new PadreID(part));
			}
		}

		return result;
	}

	private URI createRequestURI(final String methodPathKey, final Object... replacements)
	{
		String path = String.format(_methodPaths.getProperty(methodPathKey), replacements);
		return URI.create(_padreAPI.toString() + path);
	}
}
