/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PaDRe.
 *
 *	PaDRe is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PaDRe is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PaDRe.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.padre.request;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.codec.digest.DigestUtils;

/**
 * Encapsulates ID, content, and checksum of a PaDRe object.
 * @author Fabian Körner
 */
public class PadreObject
{

	/** Pattern of valid string representations of a PaDRe object with ID, content, and checksum. */
	public static final Pattern PATTERN_COMBINED =
			Pattern.compile("(([a-z]{5})[.]([0-9a-f]{3})[.]([0-9a-f]{3})[.]([0-9a-f]{9}))[#]([a-z0-9+/]+={0,2})[#]([0-9a-f]{64})", Pattern.CASE_INSENSITIVE);

	private PadreID _id;
	private String _payload;
	private String _sha256Hex;
	private String _combined;

	/**
	 * Constructor with combined components.
	 * @param combined Combined components.
	 * @throws PadreObjectException Could not create PaDRe object.
	 */
	public PadreObject(final String combined) throws PadreObjectException
	{
		_combined = combined;
		this.parseCombined();
		this.validateSHA256Hex();
	}

	/**
	 * Constructor with all possible components.
	 * @param id PaDRe object ID.
	 * @param payload PaDRe object content.
	 * @param sha256hex Checksum of object content.
	 * @throws PadreObjectException Could not create PaDRe object.
	 */
	public PadreObject(final PadreID id, final String payload, final String sha256hex) throws PadreObjectException
	{
		_id = id;
		_payload = payload;
		_sha256Hex = sha256hex;
		this.setCombined();
		this.validateSHA256Hex();
	}

	/**
	 * Returns PaDRe ID.
	 * @return PaDRe ID.
	 */
	public final PadreID getId()
	{
		return _id;
	}

	/**
	 * Returns content of PaDRe object.
	 * @return Content.
	 */
	public final String getPayload()
	{
		return _payload;
	}

	/**
	 * Returns checksum of object's content.
	 * @return Checksum.
	 */
	public final String getSHA256Hex()
	{
		return _sha256Hex;
	}

	/**
	 * Returns the combined respresentation of an object.
	 * @return All components combined.
	 */
	public final String getCombined()
	{
		return _combined;
	}

	/**
	 * Returns the pattern of valid combined representations.
	 * @return Pattern of combined form.
	 */
	public static Pattern getPatternCombined()
	{
		return PATTERN_COMBINED;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	@Override
	public final String toString()
	{
		return _combined;
	}

	private void parseCombined() throws PadreObjectException
	{
		PadreObject.validateCombined(_combined);
		String[] parts = _combined.split("#");
		try
		{
			_id = new PadreID(parts[0]);
		}
		catch (PadreIDException e)
		{
			throw new PadreObjectException(e);
		}
		_payload = parts[1];
		_sha256Hex = parts[2];
	}

	private void setCombined() throws PadreObjectException
	{
		StringBuilder sb = new StringBuilder();
		sb.append(_id.toString());
		sb.append('#');
		sb.append(_payload);
		sb.append('#');
		sb.append(_sha256Hex);
		String combined = sb.toString();

		PadreObject.validateCombined(combined);

		_combined = combined;
	}

	/**
	 * Checks if the checksum of the object is valid.
	 * @throws PadreObjectException Object has an invalid checksum.
	 */
	public final void validateSHA256Hex() throws PadreObjectException
	{
		String tmpSHA256Hex = DigestUtils.sha256Hex(_payload);
		if (_sha256Hex.equals(tmpSHA256Hex))
		{
			return;
		}
		else
		{
			throw new PadreObjectException(
					_id.toString() + ": detected invalid sha256 checksum [" + tmpSHA256Hex + "] vs. [" + _sha256Hex + "] for payload [" + _payload + "]");
		}
	}

	/**
	 * Checks the pattern and ID of the provided combined
	 * object representation.
	 * @param combined Combined components.
	 * @throws PadreObjectException Combined representation doesn't match pattern, or object has invalid ID.
	 */
	public static void validateCombined(final String combined) throws PadreObjectException
	{
		Matcher m = PATTERN_COMBINED.matcher(combined);
		if (m.find())
		{
			String padreID = null;
			try
			{
				padreID = m.group(1);
				new PadreID(padreID);
			}
			catch (PadreIDException e)
			{
				throw new PadreObjectException("PaDRe object with invalid ID: " + padreID, e);
			}
		}
		else
		{
			throw new PadreObjectException("PaDRe object [" + combined + "] doesn't fit pattern (" + PATTERN_COMBINED.toString() + ")");
		}
	}
}
