/*
 *	Copyright 2012-2014 Berlin-Brandenburgische Akademie der Wissenschaften
 *
 *	This file is part of PDR-LIB.
 *
 *	PDR-LIB is free software: you can redistribute it and/or modify it under
 *	the terms of the GNU Lesser General Public License as published by the
 *	Free Software Foundation, either version 3 of the License, or (at your
 *	option) any later version.
 *
 *	PDR-LIB is distributed in the hope that it will be useful, but WITHOUT ANY
 *	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *	FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 *	more details.
 *
 *	You should have received a copy of the GNU Lesser General Public License
 *	along with PDR-LIB.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.bbaw.pdr.solr;

import java.util.Iterator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.namespace.QName;

import org.apache.axiom.om.OMAttribute;
import org.apache.axiom.om.OMElement;

/**
 * Solr input document with index fields specific to aspects.
 * @author Fabian Körner
 */
public class SolrInputAspect extends SolrInputCommon
{

	private static final long serialVersionUID = 8537033123596655850L;

	private static final Pattern REBUILD_NOTIFICATION = Pattern.compile("[<].+?[>]");
	private static final Pattern REBUILD_NOTIFICATION_LB = Pattern.compile(
														"([<]lb[ ]?[/][>])|([<]lb[>].*?[<][/]lb[>])");

	private static final String NAME_ELEMENT_LB = "lb";

	private QName _qNameTimeDimElement;
	private QName _qNameTimeStmElement;
	private QName _qNameTimeElement;
	private QName _qNameSpatialDimElement;
	private QName _qNameSpatialStmElement;
	private QName _qNamePlaceElement;
	private QName _qNameRelationDimElement;
	private QName _qNameRelationStmElement;
	private QName _qNameRelationElement;
	private QName _qNameSemanticDimElement;
	private QName _qNameSemanticStmElement;
	private QName _qNameNotificationElement;
	private QName _qNameValidationElement;
	private QName _qNameValidationStmElement;
	private QName _qNameReferenceElement;

	private QName _qNameTypeAttribute;
	private QName _qNameSubjectAttribute;
	private QName _qNameProviderAttribute;

	protected SolrInputAspect(final OMElement root)
	throws SolrCommandException
	{
		super(root);
		this.initQNames();
		this.addSpecificIndexFields();
	}

	private void initQNames()
	{
		_qNameTimeDimElement = super.createQNameOfElement("timeDim");
		_qNameTimeStmElement = super.createQNameOfElement("timeStm");
		_qNameTimeElement = super.createQNameOfElement("time");
		_qNameSpatialDimElement = super.createQNameOfElement("spatialDim");
		_qNameSpatialStmElement = super.createQNameOfElement("spatialStm");
		_qNamePlaceElement = super.createQNameOfElement("place");
		_qNameRelationDimElement = super.createQNameOfElement("relationDim");
		_qNameRelationStmElement = super.createQNameOfElement("relationStm");
		_qNameRelationElement = super.createQNameOfElement("relation");
		_qNameSemanticDimElement = super.createQNameOfElement("semanticDim");
		_qNameSemanticStmElement = super.createQNameOfElement("semanticStm");
		_qNameNotificationElement = super.createQNameOfElement("notification");
		_qNameValidationElement = super.createQNameOfElement("validation");
		_qNameValidationStmElement = super.createQNameOfElement("validationStm");
		_qNameReferenceElement = super.createQNameOfElement("reference");

		_qNameTypeAttribute = super.createQNameOfAttribute("type");
		_qNameSubjectAttribute = super.createQNameOfAttribute("subject");
		_qNameProviderAttribute = super.createQNameOfAttribute("provider");
	}

	private void addSpecificIndexFields() throws SolrCommandException
	{
		addIndexFieldsFromTimeDimElement(super.getRoot().getFirstChildWithName(_qNameTimeDimElement));
		addIndexFieldsFromSpatialDimElement(super.getRoot().getFirstChildWithName(_qNameSpatialDimElement));
		addIndexFieldsFromRelationDimElement(super.getRoot().getFirstChildWithName(_qNameRelationDimElement));
		addIndexFieldsFromSemanticDimElement(super.getRoot().getFirstChildWithName(_qNameSemanticDimElement));
		addIndexFieldsFromNotificationElement(super.getRoot().getFirstChildWithName(_qNameNotificationElement));
		addIndexFieldsFromValidationElement(super.getRoot().getFirstChildWithName(_qNameValidationElement));
	}

	private void addIndexFieldsFromTimeDimElement(final OMElement timeDimElement) throws SolrCommandException
	{
		Iterator<?> timeStms = timeDimElement.getChildrenWithName(_qNameTimeStmElement);
		OMElement timeStmElement;
		while (timeStms.hasNext())
		{
			timeStmElement = (OMElement) timeStms.next();
			addIndexFieldsFromTimeStmElement(timeStmElement);
		}
	}

	private void addIndexFieldsFromSpatialDimElement(final OMElement spatialDimElement)
	{
		Iterator<?> spatialStms = spatialDimElement.getChildrenWithName(_qNameSpatialStmElement);
		OMElement spatialStmElement;
		while (spatialStms.hasNext())
		{
			spatialStmElement = (OMElement) spatialStms.next();
			addIndexFieldsFromSpatialStmElement(spatialStmElement);
		}
	}

	private void addIndexFieldsFromRelationDimElement(final OMElement relationDimElement)
	{
		Iterator<?> relationStms = relationDimElement.getChildrenWithName(_qNameRelationStmElement);
		OMElement relationStmElement;
		while (relationStms.hasNext())
		{
			relationStmElement = (OMElement) relationStms.next();
			addIndexFieldsFromRelationStmElement(relationStmElement);
		}
	}

	private void addIndexFieldsFromSemanticDimElement(final OMElement semanticDimElement)
	{
		Iterator<?> semanticStms = semanticDimElement.getChildrenWithName(_qNameSemanticStmElement);
		OMElement semanticStmElement;
		while (semanticStms.hasNext())
		{
			semanticStmElement = (OMElement) semanticStms.next();
			addIndexFieldsFromSemanticStmElement(semanticStmElement);
		}
	}

	private void addIndexFieldsFromNotificationElement(final OMElement notificationElement) throws SolrCommandException
	{
		String notification = SolrInputAspect.rebuildTextWithinNotification(notificationElement);
		addField("notification", notification);

		addIndexFieldsFromNotificationChildElements(notificationElement);
	}

	private void addIndexFieldsFromNotificationChildElements(final OMElement notificationElement)
	{
		Iterator<?> notificationChildElements = notificationElement.getChildElements();
		OMElement currentChild;
		while (notificationChildElements.hasNext())
		{
			currentChild = (OMElement) notificationChildElements.next();

			if (NAME_ELEMENT_LB.equals(currentChild.getLocalName()))
			{
				/* ignore <lb /> */
			}
			else
			{
				SolrInputAspect.NotificationChild notificationChild;
				notificationChild = new SolrInputAspect.NotificationChild(currentChild);
				addIndexFieldsFromNotificationChild(notificationChild);
			}
		}
	}

	private void addIndexFieldsFromValidationElement(final OMElement validationElement)
	{
		Iterator<?> validationStms = validationElement.getChildrenWithName(_qNameValidationStmElement);
		OMElement validationStmElement;
		while (validationStms.hasNext())
		{
			validationStmElement = (OMElement) validationStms.next();
			addIndexFieldsFromValidationStmElement(validationStmElement);
		}
	}

	private void addIndexFieldsFromTimeStmElement(final OMElement timeStmElement) throws SolrCommandException
	{
		Iterator<?> times = timeStmElement.getChildrenWithName(_qNameTimeElement);
		OMElement timeElement;
		while (times.hasNext())
		{
			timeElement = (OMElement) times.next();
			addIndexFieldsFromTimeElement(timeElement);
		}
	}

	private void addIndexFieldsFromSpatialStmElement(final OMElement spatialStmElement)
	{
		Iterator<?> places = spatialStmElement.getChildrenWithName(_qNamePlaceElement);
		SolrInputAspect.Place aspectPlace;
		while (places.hasNext())
		{
			aspectPlace = new SolrInputAspect.Place((OMElement) places.next());
			addIndexFieldsFromAspectPlace(aspectPlace);
		}
	}

	private void addIndexFieldsFromRelationStmElement(final OMElement relationStmElement)
	{
		String subject = relationStmElement.getAttributeValue(_qNameSubjectAttribute);
		Iterator<?> relations = relationStmElement.getChildrenWithName(_qNameRelationElement);
		SolrInputAspect.Relation aspectRelation;
		while (relations.hasNext())
		{
			aspectRelation = new SolrInputAspect.Relation(subject, (OMElement) relations.next());
			addIndexFieldsFromAspectRelation(aspectRelation);
		}
	}

	private void addIndexFieldsFromSemanticStmElement(final OMElement semanticStmElement)
	{
		String semantic = semanticStmElement.getText();
		String provider = semanticStmElement.getAttributeValue(_qNameProviderAttribute);
		String csemantic = SolrInputCommon.getCombinedIndexValue(semantic, provider);

		addField("semantic", semantic);
		addField("semantic.provider", provider);
		addField("csemantic", csemantic);
	}

	private void addIndexFieldsFromValidationStmElement(final OMElement validationStmElement)
	{
		Iterator<?> references = validationStmElement.getChildrenWithName(_qNameReferenceElement);
		OMElement referenceElement;
		while (references.hasNext())
		{
			referenceElement = (OMElement) references.next();
			addField("validation.reference", referenceElement.getText());
		}
	}

	private void addIndexFieldsFromTimeElement(final OMElement timeElement) throws SolrCommandException
	{
		OMAttribute typeAttribute = timeElement.getAttribute(_qNameTypeAttribute);
		String time = timeElement.getText();
		String type = typeAttribute.getAttributeValue();
		String ctime = SolrInputCommon.getCombinedIndexValue(time, type);

		String solrTime = SolrCompliantTime.createFromFragment(time);

		addField("time", solrTime);
		addField("time.type", type);
		addField("ctime", ctime);
	}

	private void addIndexFieldsFromAspectPlace(final SolrInputAspect.Place aspectPlace)
	{
		String place = aspectPlace.getPlaceValue();
		String type = aspectPlace.getTypeValue();
		String subtype = aspectPlace.getSubtypeValue();
		String key = aspectPlace.getKeyValue();
		String cplace = SolrInputCommon.getCombinedIndexValue(place, type,
				(aspectPlace.hasSubtype() ? subtype : "null"),
				(aspectPlace.hasKey() ? key : "null")
				);

		addField("place", place);
		addField("place.type", type);
		if (aspectPlace.hasSubtype())
		{
			addField("place.subtype", subtype);
		}
		if (aspectPlace.hasKey())
		{
			addField("place.key", key);
		}
		addField("cplace", cplace);
	}

	private void addIndexFieldsFromAspectRelation(final SolrInputAspect.Relation aspectRelation)
	{
		String relation = aspectRelation.getRelationValue();
		String subject = aspectRelation.getSubjectValue();
		String object = aspectRelation.getObjectValue();
		String provider = aspectRelation.getProviderValue();
		String clazz = aspectRelation.getClassValue();
		String context = aspectRelation.getContextValue();
		String crelation = SolrInputCommon.getCombinedIndexValue(subject, relation, object, provider,
				(aspectRelation.hasClass() ? clazz : "null"), (aspectRelation.hasContext() ? context : "null"));

		addField("relation", relation);
		addField("relation.subject", subject);
		addField("relation.object", object);
		addField("relation.provider", provider);
		if (aspectRelation.hasClass())
		{
			addField("relation.class", clazz);
		}
		if (aspectRelation.hasContext())
		{
			addField("relation.context", context);
		}
		addField("crelation", crelation);
	}

	private void addIndexFieldsFromNotificationChild(final SolrInputAspect.NotificationChild notificationChild)
	{
		String elementNameLowerCase = notificationChild.getElementName().toLowerCase();
		String elementValue = notificationChild.getElementValue();
		String type = notificationChild.getTypeValue();
		String subtype = notificationChild.getSubtypeValue();
		String role = notificationChild.getRoleValue();
		String key = notificationChild.getKeyValue();
		String ana = notificationChild.getAnaValue();
		String cValue = SolrInputCommon.getCombinedIndexValue(elementValue, type,
				(notificationChild.hasSubtype() ? subtype : "null"), (notificationChild.hasRole() ? role : "null"),
				(notificationChild.hasKey() ? key : "null"), (notificationChild.hasAna() ? ana : "null")
				);
		addField("notification." + elementNameLowerCase, elementValue);
		addField("notification." + elementNameLowerCase + ".type", type);
		if (notificationChild.hasSubtype())
		{
			addField("notification." + elementNameLowerCase + ".subtype", subtype);
		}
		if (notificationChild.hasRole())
		{
			addField("notification." + elementNameLowerCase + ".role", role);
		}
		if (notificationChild.hasKey())
		{
			addField("notification." + elementNameLowerCase + ".key", key);
		}
		if (notificationChild.hasAna())
		{
			addField("notification." + elementNameLowerCase + ".ana", ana);
		}
		addField("notification.c" + elementNameLowerCase, cValue);
	}

	protected static String rebuildTextWithinNotification(final OMElement element)
	{
		String result = element.toString();
		Matcher matcher;

		matcher = REBUILD_NOTIFICATION_LB.matcher(result);
		result = matcher.replaceAll("\n");
		matcher = REBUILD_NOTIFICATION.matcher(result);
		result = matcher.replaceAll("");

		return result;
	}

	/**
	 *
	 * @author Fabian Körner
	 */
	private static final class Place
	{
		private static final QName QNAME_TYPE_ATTRIBUTE = new QName("type");
		private static final QName QNAME_SUBTYPE_ATTRIBUTE = new QName("subtype");
		private static final QName QNAME_KEY_ATTRIBUTE = new QName("key");

		private OMElement _placeElement;

		private String _placeValue;
		private String _typeValue;
		private String _subtypeValue;
		private String _keyValue;

		private boolean _hasSubtype;
		private boolean _hasKey;

		private Place(final OMElement placeElement)
		{
			_placeElement = placeElement;
			this.setPlaceValue();
			this.setTypeValue();
			this.setSubtypeValue();
			this.setKeyValue();
		}

		private void setPlaceValue()
		{
			_placeValue = _placeElement.getText();
		}

		private void setTypeValue()
		{
			OMAttribute typeAttribute = _placeElement.getAttribute(QNAME_TYPE_ATTRIBUTE);
			_typeValue = typeAttribute.getAttributeValue();
		}

		private void setSubtypeValue()
		{
			OMAttribute subtypeAttribute = _placeElement.getAttribute(QNAME_SUBTYPE_ATTRIBUTE);
			if (null != subtypeAttribute)
			{
				_subtypeValue = subtypeAttribute.getAttributeValue();
				if (0 == _subtypeValue.length())
				{
					_hasSubtype = false;
				}
				else
				{
					_hasSubtype = true;
				}
			}
			else
			{
				_hasSubtype = false;
				_subtypeValue = "";
			}
		}

		private void setKeyValue()
		{
			OMAttribute keyAttribute = _placeElement.getAttribute(QNAME_KEY_ATTRIBUTE);
			if (null != keyAttribute)
			{
				_keyValue = keyAttribute.getAttributeValue();
				if (0 == _keyValue.length())
				{
					_hasKey = false;
				}
				else
				{
					_hasKey = true;
				}
			}
			else
			{
				_hasKey = false;
				_keyValue = "";
			}
		}

		private String getPlaceValue()
		{
			return _placeValue;
		}

		private String getTypeValue()
		{
			return _typeValue;
		}

		private String getSubtypeValue()
		{
			return _subtypeValue;
		}

		private String getKeyValue()
		{
			return _keyValue;
		}

		private boolean hasSubtype()
		{
			return _hasSubtype;
		}

		private boolean hasKey()
		{
			return _hasKey;
		}
	}

	/**
	 *
	 * @author Fabian Körner
	 */
	private static final class Relation
	{
		private static final QName QNAME_OBJECT_ATTRIBUTE = new QName("object");
		private static final QName QNAME_PROVIDER_ATTRIBUTE = new QName("provider");
		private static final QName QNAME_CLASS_ATTRIBUTE = new QName("class");
		private static final QName QNAME_CONTEXT_ATTRIBUTE = new QName("context");

		private OMElement _relationElement;

		private String _subjectValue;
		private String _relationValue;
		private String _objectValue;
		private String _providerValue;
		private String _classValue;
		private String _contextValue;

		private boolean _hasClass;
		private boolean _hasContext;

		private Relation(final String subjectValue, final OMElement relationElement)
		{
			_subjectValue = subjectValue;
			_relationElement = relationElement;
			this.setRelationValue();
			this.setObjectValue();
			this.setProviderValue();
			this.setClassValue();
			this.setContextValue();
		}

		private void setRelationValue()
		{
			_relationValue = _relationElement.getText();
		}

		private void setObjectValue()
		{
			OMAttribute objectAttribute = _relationElement.getAttribute(QNAME_OBJECT_ATTRIBUTE);
			_objectValue = objectAttribute.getAttributeValue();
		}

		private void setProviderValue()
		{
			OMAttribute providerAttribute = _relationElement.getAttribute(QNAME_PROVIDER_ATTRIBUTE);
			_providerValue = providerAttribute.getAttributeValue();
		}

		private void setClassValue()
		{
			OMAttribute classAttribute = _relationElement.getAttribute(QNAME_CLASS_ATTRIBUTE);
			if (null != classAttribute)
			{
				_classValue = classAttribute.getAttributeValue();
				if (0 == _classValue.length())
				{
					_hasClass = false;
				}
				else
				{
					_hasClass = true;
				}
			}
			else
			{
				_hasClass = false;
				_classValue = "";
			}
		}

		private void setContextValue()
		{
			OMAttribute contextAttribute = _relationElement.getAttribute(QNAME_CONTEXT_ATTRIBUTE);
			if (null != contextAttribute)
			{
				_contextValue = contextAttribute.getAttributeValue();
				if (0 == _contextValue.length())
				{
					_hasContext = false;
				}
				else
				{
					_hasContext = true;
				}
			}
			else
			{
				_hasContext = false;
				_contextValue = "";
			}
		}

		private String getSubjectValue()
		{
			return _subjectValue;
		}

		private String getRelationValue()
		{
			return _relationValue;
		}

		private String getObjectValue()
		{
			return _objectValue;
		}

		private String getProviderValue()
		{
			return _providerValue;
		}

		private String getClassValue()
		{
			return _classValue;
		}

		private String getContextValue()
		{
			return _contextValue;
		}

		private boolean hasClass()
		{
			return _hasClass;
		}

		private boolean hasContext()
		{
			return _hasContext;
		}
	}

	/**
	 *
	 * @author Fabian Körner
	 */
	private static final class NotificationChild
	{
		private static final QName QNAME_TYPE_ATTRIBUTE = new QName("type");
		private static final QName QNAME_SUBTYPE_ATTRIBUTE = new QName("subtype");
		private static final QName QNAME_ROLE_ATTRIBUTE = new QName("role");
		private static final QName QNAME_KEY_ATTRIBUTE = new QName("key");
		private static final QName QNAME_ANA_ATTRIBUTE = new QName("ana");

		private OMElement _notificationChildElement;

		private String _elementName;
		private String _elementValue;
		private String _typeValue;
		private String _subtypeValue;
		private String _roleValue;
		private String _keyValue;
		private String _anaValue;

		private boolean _hasSubtype;
		private boolean _hasRole;
		private boolean _hasKey;
		private boolean _hasAna;

		private NotificationChild(final OMElement notificationChildElement)
		{
			_notificationChildElement = notificationChildElement;
			_elementName = _notificationChildElement.getLocalName();
			_elementValue = SolrInputAspect.rebuildTextWithinNotification(_notificationChildElement);
			_typeValue = _notificationChildElement.getAttributeValue(QNAME_TYPE_ATTRIBUTE);
			this.setSubtypeValue();
			this.setRoleValue();
			this.setKeyValue();
			this.setAnaValue();
		}

		private void setSubtypeValue()
		{
			OMAttribute subtypeAttribute = _notificationChildElement.getAttribute(QNAME_SUBTYPE_ATTRIBUTE);
			if (null != subtypeAttribute)
			{
				_subtypeValue = subtypeAttribute.getAttributeValue();
				if (0 == _subtypeValue.length())
				{
					_hasSubtype = false;
				}
				else
				{
					_hasSubtype = true;
				}
			}
			else
			{
				_hasSubtype = false;
				_subtypeValue = "";
			}
		}

		private void setRoleValue()
		{
			OMAttribute roleAttribute = _notificationChildElement.getAttribute(QNAME_ROLE_ATTRIBUTE);
			if (null != roleAttribute)
			{
				_roleValue = roleAttribute.getAttributeValue();
				if (0 == _roleValue.length())
				{
					_hasRole = false;
				}
				else
				{
					_hasRole = true;
				}
			}
			else
			{
				_hasRole = false;
				_roleValue = "";
			}
		}

		private void setKeyValue()
		{
			OMAttribute keyAttribute = _notificationChildElement.getAttribute(QNAME_KEY_ATTRIBUTE);
			if (null != keyAttribute)
			{
				_keyValue = keyAttribute.getAttributeValue();
				if (0 == _keyValue.length())
				{
					_hasKey = false;
				}
				else
				{
					_hasKey = true;
				}
			}
			else
			{
				_hasKey = false;
				_keyValue = "";
			}
		}

		private void setAnaValue()
		{
			OMAttribute anaAttribute = _notificationChildElement.getAttribute(QNAME_ANA_ATTRIBUTE);
			if (null != anaAttribute)
			{
				_anaValue = anaAttribute.getAttributeValue();
				if (0 == _anaValue.length())
				{
					_hasAna = false;
				}
				else
				{
					_hasAna = true;
				}
			}
			else
			{
				_hasAna = false;
				_anaValue = "";
			}
		}

		private String getElementName()
		{
			return _elementName;
		}

		private String getElementValue()
		{
			return _elementValue;
		}

		private String getTypeValue()
		{
			return _typeValue;
		}

		private String getSubtypeValue()
		{
			return _subtypeValue;
		}

		private String getRoleValue()
		{
			return _roleValue;
		}

		private String getKeyValue()
		{
			return _keyValue;
		}

		private String getAnaValue()
		{
			return _anaValue;
		}

		private boolean hasSubtype()
		{
			return _hasSubtype;
		}

		private boolean hasRole()
		{
			return _hasRole;
		}

		private boolean hasKey()
		{
			return _hasKey;
		}

		private boolean hasAna()
		{
			return _hasAna;
		}
	}
}
